@@ -4,12 +4,12 @@ import numba
 from sklearn.base import BaseEstimator, TransformerMixin
 from sklearn.utils import check_array, check_random_state
 from sklearn.utils.extmath import randomized_svd
+from sklearn.utils.validation import _check_sample_weight
 from sklearn.decomposition import non_negative_factorization
 from scipy.sparse import issparse, csr_matrix, coo_matrix
 
 from enstop.utils import normalize, coherence, mean_coherence, log_lift, mean_log_lift
 
-
 @numba.njit(
     "f4[:,::1](i4[::1],i4[::1],f4[::1],f4[:,::1],f4[:,::1],f4[:,::1],f4)",
     locals={
@@ -110,7 +110,14 @@ def plsa_e_step(
     parallel=True,
 )
 def plsa_m_step(
-    X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d, p_z_given_wd, norm_pwz, norm_pdz
+    X_rows,
+    X_cols,
+    X_vals,
+    p_w_given_z,
+    p_z_given_d,
+    p_z_given_wd,
+    norm_pwz,
+    norm_pdz
 ):
     """Perform the M-step of pLSA optimization. This amounts to using the estimates
     of P(z|w,d) to estimate the values P(w|z) and P(z|d). The computation implements
@@ -145,6 +152,9 @@ def plsa_m_step(
     p_z_given_wd: array of shape (nnz, n_topics)
         The current estimates for P(z|w,d)
 
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
     norm_pwz: array of shape (n_topics,)
         Auxilliary array used for storing row norms; this is passed in to save
         reallocations.
@@ -152,7 +162,6 @@ def plsa_m_step(
     norm_pdz: array of shape (n_docs,)
         Auxilliary array used for storing row norms; this is passed in to save
         reallocations.
-
     """
 
     k = p_z_given_wd.shape[1]
@@ -191,7 +200,113 @@ def plsa_m_step(
 
 
 @numba.njit(
-    "f4(i4[::1],i4[::1],f4[::1],f4[:,::1],f4[:,::1])",
+    "UniTuple(f4[:,::1],2)(i4[::1],i4[::1],f4[::1],f4[:,::1],f4[:,::1],f4[:,::1],f4[::1],f4[::1],f4[::1])",
+    locals={
+        "k": numba.types.uint16,
+        "w": numba.types.uint32,
+        "d": numba.types.uint32,
+        "z": numba.types.uint16,
+        "nz_idx": numba.types.uint32,
+        "s": numba.types.float32,
+    },
+    fastmath=True,
+    nogil=True,
+    parallel=True,
+)
+def plsa_m_step_w_sample_weight(
+    X_rows,
+    X_cols,
+    X_vals,
+    p_w_given_z,
+    p_z_given_d,
+    p_z_given_wd,
+    sample_weight,
+    norm_pwz,
+    norm_pdz
+):
+    """Perform the M-step of pLSA optimization. This amounts to using the estimates
+    of P(z|w,d) to estimate the values P(w|z) and P(z|d). The computation implements
+
+    P(w|z) = \frac{\sum_{d\in D} X_{w,d}P(z|w,d)}{\sum_{d,z} X_{w,d}P(z|w,d)}
+    P(z|d) = \frac{\sum_{w\in V} X_{w,d}P(z|w,d)}{\sum_{w,d} X_{w,d}P(z|w,d)}
+
+    This routine is optimized to work with sparse matrices such that P(z|w,d) is only
+    computed for w, d such that X_{w,d} is non-zero, where X is the data matrix.
+
+    To make this numba compilable the raw arrays defining the COO format sparse
+    matrix must be passed separately.
+
+    Parameters
+    ----------
+    X_rows: array of shape (nnz,)
+        For each non-zero entry of X, the row of the entry.
+
+    X_cols: array of shape (nnz,)
+        For each non-zero entry of X, the column of the
+        entry.
+
+    X_vals: array of shape (nnz,)
+        For each non-zero entry of X, the value of entry.
+
+    p_w_given_z: array of shape (n_topics, n_words)
+        The result array to write new estimates of P(w|z) to.
+
+    p_z_given_d: array of shape (n_docs, n_topics)
+        The result array to write new estimates of P(z|d) to.
+
+    p_z_given_wd: array of shape (nnz, n_topics)
+        The current estimates for P(z|w,d)
+
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
+    norm_pwz: array of shape (n_topics,)
+        Auxilliary array used for storing row norms; this is passed in to save
+        reallocations.
+
+    norm_pdz: array of shape (n_docs,)
+        Auxilliary array used for storing row norms; this is passed in to save
+        reallocations.
+    """
+
+    k = p_z_given_wd.shape[1]
+    n = p_z_given_d.shape[0]
+    m = p_w_given_z.shape[1]
+
+    p_w_given_z[:] = 0.0
+    p_z_given_d[:] = 0.0
+
+    norm_pwz[:] = 0.0
+    norm_pdz[:] = 0.0
+
+    for nz_idx in range(X_vals.shape[0]):
+        d = X_rows[nz_idx]
+        w = X_cols[nz_idx]
+        x = X_vals[nz_idx]
+
+        for z in range(k):
+            s = x * p_z_given_wd[nz_idx, z]
+            t = s * sample_weight[d]
+
+            p_w_given_z[z, w] += t
+            p_z_given_d[d, z] += s
+
+            norm_pwz[z] += t
+            norm_pdz[d] += s
+
+    for z in numba.prange(k):
+        if norm_pwz[z] > 0:
+            for w in range(m):
+                p_w_given_z[z, w] /= norm_pwz[z]
+        for d in range(n):
+            if norm_pdz[d] > 0:
+                p_z_given_d[d, z] /= norm_pdz[d]
+
+    return p_w_given_z, p_z_given_d
+
+
+@numba.njit(
+    "f4(i4[::1],i4[::1],f4[::1],f4[:,::1],f4[:,::1],f4[::1])",
     locals={
         "k": numba.types.uint16,
         "w": numba.types.uint32,
@@ -206,7 +321,7 @@ def plsa_m_step(
     nogil=True,
     parallel=True,
 )
-def log_likelihood(X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d):
+def log_likelihood(X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d, sample_weight):
     """Compute the log-likelihood of observing the data X given estimates for P(w|z)
     and P(z|d). The likelihood of X_{w,d} under the model is given by X_{w,d} P(w|d)
     = X_{w,d} P(w|z) P(z|d). This function returns
@@ -237,6 +352,9 @@ def log_likelihood(X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d):
     p_z_given_d: array of shape (n_docs, n_topics)
         The current estimates of values for P(z|d)
 
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
     Returns
     -------
 
@@ -258,7 +376,7 @@ def log_likelihood(X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d):
         for z in range(k):
             p_w_given_d += p_w_given_z[z, w] * p_z_given_d[d, z]
 
-        result += x * np.log(p_w_given_d)
+        result += x * np.log(p_w_given_d) * sample_weight[d]
 
     return result
 
@@ -397,10 +515,12 @@ def plsa_fit_inner(
     X_vals,
     p_w_given_z,
     p_z_given_d,
+    sample_weight,
     n_iter=100,
     n_iter_per_test=10,
     tolerance=0.001,
     e_step_thresh=1e-32,
+    use_sample_weights=False,
 ):
     """Internal loop of EM steps required to optimize pLSA, along with relative
     convergence tests with respect to the log-likelihood of observing the data under
@@ -431,6 +551,9 @@ def plsa_fit_inner(
     p_z_given_d: array of shape (n_docs, n_topics)
         The current estimates of values for P(z|d)
 
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
     n_iter: int
         The maximum number iterations of EM to perform
 
@@ -461,7 +584,7 @@ def plsa_fit_inner(
     norm_pdz = np.zeros(n, dtype=np.float32)
 
     previous_log_likelihood = log_likelihood(
-        X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d
+        X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d, sample_weight
     )
 
     for i in range(n_iter):
@@ -475,20 +598,33 @@ def plsa_fit_inner(
             p_z_given_wd,
             e_step_thresh,
         )
-        plsa_m_step(
-            X_rows,
-            X_cols,
-            X_vals,
-            p_w_given_z,
-            p_z_given_d,
-            p_z_given_wd,
-            norm_pwz,
-            norm_pdz,
-        )
+        if use_sample_weights:
+            plsa_m_step_w_sample_weight(
+                X_rows,
+                X_cols,
+                X_vals,
+                p_w_given_z,
+                p_z_given_d,
+                p_z_given_wd,
+                sample_weight,
+                norm_pwz,
+                norm_pdz,
+            )
+        else:
+            plsa_m_step(
+                X_rows,
+                X_cols,
+                X_vals,
+                p_w_given_z,
+                p_z_given_d,
+                p_z_given_wd,
+                norm_pwz,
+                norm_pdz,
+            )
 
         if i % n_iter_per_test == 0:
             current_log_likelihood = log_likelihood(
-                X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d
+                X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d, sample_weight
             )
             change = np.abs(current_log_likelihood - previous_log_likelihood)
             if change / np.abs(current_log_likelihood) < tolerance:
@@ -502,6 +638,7 @@ def plsa_fit_inner(
 def plsa_fit(
     X,
     k,
+    sample_weight,
     init="random",
     n_iter=100,
     n_iter_per_test=10,
@@ -524,6 +661,9 @@ def plsa_fit(
     k: int
         The number of topics for pLSA to fit with.
 
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
     init: string or tuple (optional, default="random")
         The intialization method to use. This should be one of:
             * ``"random"``
@@ -564,6 +704,8 @@ def plsa_fit(
     p_z_given_d = p_z_given_d.astype(np.float32, order="C")
     p_w_given_z = p_w_given_z.astype(np.float32, order="C")
 
+    use_sample_weights = np.any(sample_weight != 1.0)
+
     A = X.tocoo().astype(np.float32)
 
     p_z_given_d, p_w_given_z = plsa_fit_inner(
@@ -572,17 +714,19 @@ def plsa_fit(
         A.data,
         p_w_given_z,
         p_z_given_d,
+        sample_weight,
         n_iter,
         n_iter_per_test,
         tolerance,
         e_step_thresh,
+        use_sample_weights,
     )
 
     return p_z_given_d, p_w_given_z
 
 
 @numba.njit(
-    "UniTuple(f4[:,::1],2)(i4[::1],i4[::1],f4[::1],f4[:,::1],f4[:,::1],f4[:,::1],f4[::1])",
+    "UniTuple(f4[:,::1],2)(i4[::1],i4[::1],f4[::1],f4[:,::1],f4[:,::1],f4[:,::1],f4[::1],f4[::1])",
     locals={
         "k": numba.types.uint16,
         "w": numba.types.uint32,
@@ -595,7 +739,14 @@ def plsa_fit(
     nogil=True,
 )
 def plsa_refit_m_step(
-    X_rows, X_cols, X_vals, p_w_given_z, p_z_given_d, p_z_given_wd, norm_pdz
+    X_rows,
+    X_cols,
+    X_vals,
+    p_w_given_z,
+    p_z_given_d,
+    p_z_given_wd,
+    sample_weight,
+    norm_pdz
 ):
     """Optimized routine for the M step fitting values of P(z|d) given a fixed set of
     topics (i.e. P(w|z)).
@@ -627,6 +778,9 @@ def plsa_refit_m_step(
     p_z_given_wd: array of shape (nnz, n_topics)
         The current estimates for P(z|w,d)
 
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
     norm_pdz: array of shape (n_docs,)
         Auxilliary array used for storing row norms; this is passed in to save
         reallocations.
@@ -664,6 +818,7 @@ def plsa_refit_inner(
     X_vals,
     topics,
     p_z_given_d,
+    sample_weight,
     n_iter=50,
     n_iter_per_test=10,
     tolerance=0.005,
@@ -697,6 +852,9 @@ def plsa_refit_inner(
     p_z_given_d: array of shape (n_docs, n_topics)
         The current estimates of values for P(z|d)
 
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
     n_iter: int
         The maximum number iterations of EM to perform
 
@@ -724,7 +882,7 @@ def plsa_refit_inner(
     norm_pdz = np.zeros(p_z_given_d.shape[0], dtype=np.float32)
 
     previous_log_likelihood = log_likelihood(
-        X_rows, X_cols, X_vals, topics, p_z_given_d
+        X_rows, X_cols, X_vals, topics, p_z_given_d, sample_weight
     )
 
     for i in range(n_iter):
@@ -733,12 +891,12 @@ def plsa_refit_inner(
             X_rows, X_cols, X_vals, topics, p_z_given_d, p_z_given_wd, e_step_thresh
         )
         plsa_refit_m_step(
-            X_rows, X_cols, X_vals, topics, p_z_given_d, p_z_given_wd, norm_pdz
+            X_rows, X_cols, X_vals, topics, p_z_given_d, p_z_given_wd, sample_weight, norm_pdz
         )
 
         if i % n_iter_per_test == 0:
             current_log_likelihood = log_likelihood(
-                X_rows, X_cols, X_vals, topics, p_z_given_d
+                X_rows, X_cols, X_vals, topics, p_z_given_d, sample_weight
             )
             if current_log_likelihood > 0:
                 change = np.abs(current_log_likelihood - previous_log_likelihood)
@@ -753,6 +911,7 @@ def plsa_refit_inner(
 def plsa_refit(
     X,
     topics,
+    sample_weight,
     n_iter=50,
     n_iter_per_test=10,
     tolerance=0.005,
@@ -771,6 +930,9 @@ def plsa_refit(
     topics: array of shape (n_topics, n_words)
         The fixed topics against which to fit the values of P(z|d).
 
+    sample_weight: array of shape (n_docs,)
+        Input document weights.
+
     n_iter: int
         The maximum number iterations of EM to perform
 
@@ -813,6 +975,7 @@ def plsa_refit(
         A.data,
         topics,
         p_z_given_d,
+        sample_weight,
         n_iter=n_iter,
         n_iter_per_test=n_iter_per_test,
         tolerance=tolerance,
@@ -917,7 +1080,7 @@ class PLSA(BaseEstimator, TransformerMixin):
         self.transform_random_seed = transform_random_seed
         self.random_state = random_state
 
-    def fit(self, X, y=None):
+    def fit(self, X, y=None, sample_weight=None):
         """Learn the pLSA model for the data X and return the document vectors.
 
         This is more efficient than calling fit followed by transform.
@@ -929,14 +1092,17 @@ class PLSA(BaseEstimator, TransformerMixin):
 
         y: Ignored
 
+        sample_weight: array of shape (n_docs,)
+            Input document weights.
+
         Returns
         -------
         self
         """
-        self.fit_transform(X)
+        self.fit_transform(X, sample_weight=sample_weight)
         return self
 
-    def fit_transform(self, X, y=None):
+    def fit_transform(self, X, y=None, sample_weight=None):
         """Learn the pLSA model for the data X and return the document vectors.
 
         This is more efficient than calling fit followed by transform.
@@ -948,6 +1114,9 @@ class PLSA(BaseEstimator, TransformerMixin):
 
         y: Ignored
 
+        sample_weight: array of shape (n_docs,)
+            Input document weights.
+
         Returns
         -------
         embedding: array of shape (n_docs, n_topics)
@@ -959,9 +1128,28 @@ class PLSA(BaseEstimator, TransformerMixin):
         if not issparse(X):
             X = csr_matrix(X)
 
+
+        sample_weight = _check_sample_weight(
+            sample_weight, X, dtype=np.float32)
+
+        if np.any(X.data < 0):
+            raise ValueError("PLSA is only valid for matrices with non-negative "
+                             "entries")
+
+        row_sums = np.array(X.sum(axis=1).T)[0]
+        good_rows = row_sums != 0
+
+        if not np.all(good_rows):
+            zero_rows_found = True
+            data_for_fitting = X[good_rows]
+        else:
+            zero_rows_found = False
+            data_for_fitting = X
+
         U, V = plsa_fit(
-            X,
+            data_for_fitting,
             self.n_components,
+            sample_weight,
             self.init,
             self.n_iter,
             self.n_iter_per_test,
@@ -969,8 +1157,14 @@ class PLSA(BaseEstimator, TransformerMixin):
             self.e_step_thresh,
             self.random_state,
         )
+
+        if zero_rows_found:
+            self.embedding_ = np.zeros((X.shape[0], self.n_components))
+            self.embedding_[good_rows] = U
+        else:
+            self.embedding_ = U
+
         self.components_ = V
-        self.embedding_ = U
         self.training_data_ = X
 
         return U
@@ -993,6 +1187,10 @@ class PLSA(BaseEstimator, TransformerMixin):
         X = check_array(X, accept_sparse="csr")
         random_state = check_random_state(self.transform_random_seed)
 
+        # Set weights to 1 for all examples
+        sample_weight = _check_sample_weight(
+            None, X, dtype=np.float32)
+
         if not issparse(X):
             X = coo_matrix(X)
         else:
@@ -1001,6 +1199,7 @@ class PLSA(BaseEstimator, TransformerMixin):
         result = plsa_refit(
             X,
             self.components_,
+            sample_weight,
             n_iter=50,
             n_iter_per_test=5,
             tolerance=0.001,
